# Ariane Docker plugin
# Docker exceptions
#
# Copyright (C) 2016 echinopsii
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ariane_clip3.exceptions import ArianeError

__author__ = 'mffrench'


class ArianeDockerError(ArianeError):
    def __repr__(self):
        return "Unspecified Ariane Docker Error occurred"


class ArianeDockerConfigFileError(ArianeDockerError):
    def __repr__(self):
        return self.args[0] + " is not a file !"


class ArianeDockerConfigMandatorySectionMissingError(ArianeDockerError):
    def __repr__(self):
        return self.args[0] + " is not defined in Ariane Docker config file !"


class ArianeDockerConfigMandatoryFieldsMissingError(ArianeDockerError):
    def __repr__(self):
        return '['.join(self.args[0]) + "] (is | are) not defined in Ariane Docker config file !"


class ArianeDockerConfigMandatoryFieldsValueError(ArianeDockerError):
    def __repr__(self):
        return self.args[0] + " value is not correct. Reason: " + self.args[1]


# Ariane Docker plugin
# Docker tooling from docker-py to Ariane server
#
# Copyright (C) 2016 echinopsii
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import copy
from ipaddress import IPv4Network
import logging
import os
import pprint
import subprocess
import tempfile
from sys import platform as _platform
from ariane_procos.system import MapSocket, NetworkInterfaceCard
from nsenter import Namespace

__author__ = 'mffrench'

LOGGER = logging.getLogger(__name__)

class DockerImage(object):
    pass

class DockerNetwork(object):
    def __init__(self, bridge_name=None, subnet_id=None, subnet=None, nic_id=None, nic=None,
                 nid=None, driver=None, IPAM=None, name=None, options=None, scope=None, containers_network=None):
        self.subnet_id = subnet_id
        self.subnet = subnet
        self.nic_id = nic_id
        self.nic = nic

        self.nid = nid
        self.driver = driver
        self.IPAM = IPAM
        self.name = name
        self.bridge_name = bridge_name
        self.options = options
        self.scope = scope
        self.containers_network = containers_network

    def __eq__(self, other):
        return self.nid == other.nid

    def to_json(self):
        #LOGGER.debug("DockerNetwork.to_json")
        json_obj = {
            'nid': self.nid,
            'driver': self.driver,
            'IPAM': self.IPAM,
            'name': self.name,
            'bridge_name': self.bridge_name,
            'options': self.options,
            'scope': self.scope,
            'containers_network': self.containers_network
        }
        return json_obj

    @staticmethod
    def from_json(json_obj):
        #LOGGER.debug("DockerNetwork.from_json")
        return DockerNetwork(
            nid=json_obj['nid'] if json_obj['nid'] else None,
            driver=json_obj['driver'] if json_obj['driver'] else None,
            IPAM=json_obj['IPAM'] if json_obj['IPAM'] else None,
            name=json_obj['name'] if json_obj['name'] else None,
            bridge_name=json_obj['bridge_name'] if json_obj['bridge_name'] else None,
            options=json_obj['options'] if json_obj['options'] else None,
            scope=json_obj['scope'] if json_obj['scope'] else None,
            containers_network=json_obj['containers_network'] if json_obj['containers_network'] else None
        )

class DockerContainerProcess(object):
    def __init__(self, pid=None, mdpid=None, mdp=None, mospid=None, mosp=None, mcid=None, mc=None,
                 map_sockets=None, last_map_sockets=None):
        self.pid = pid

        self.mdpid = mdpid #mapping docker process node id
        self.mdp = mdp
        self.mospid = mospid #mapping os process node id
        self.mosp = mosp
        self.mcid = mcid #mapping container id (parent)
        self.mc = mc

        self.map_sockets = map_sockets if map_sockets is not None else []
        self.last_map_sockets = last_map_sockets if last_map_sockets is not None else []
        self.new_map_sockets = []

    def __eq__(self, other):
        return self.pid == other.pid

    def to_json(self):
        #LOGGER.debug("DockerContainerProcess.to_json")
        map_socket_2_json = []
        for map_socket in self.map_sockets:
            map_socket_2_json.append(map_socket.to_json())
        last_map_socket_2_json = []
        for last_map_socket in self.last_map_sockets:
            last_map_socket_2_json.append(last_map_socket.to_json())
        json_obj = {
            'pid': self.pid,
            'mdpid': self.mdpid,
            'mospid': self.mospid,
            'mcid': self.mcid,
            'map_sockets': map_socket_2_json,
            'last_map_sockets': last_map_socket_2_json
        }
        return json_obj

    @staticmethod
    def from_json(json_obj):
        #LOGGER.debug("DockerContainerProcess.from_json")
        map_sockets_json = json_obj['map_sockets'] if json_obj['map_sockets'] else []
        map_sockets = []
        for map_socket_json in map_sockets_json:
            map_sockets.append(MapSocket.from_json(map_socket_json))
        last_map_sockets_json = json_obj['last_map_sockets'] if json_obj['last_map_sockets'] else []
        last_map_sockets = []
        for last_map_socket_json in last_map_sockets_json:
            last_map_sockets.append(MapSocket.from_json(last_map_socket_json))
        return DockerContainerProcess(
            pid=json_obj['pid'] if json_obj['pid'] else None,
            mdpid=json_obj['mdpid'] if json_obj['mdpid'] else None,
            mospid=json_obj['mospid'] if json_obj['mospid'] else None,
            mcid=json_obj['mcid'] if json_obj['mcid'] else None,
            map_sockets=map_sockets,
            last_map_sockets=last_map_sockets
        )

class DockerContainer(object):

    ariane_ost_name = "ARIANE_OS_TYPE_NAME"
    ariane_ost_arc = "ARIANE_OS_TYPE_ARCHITECTURE"
    ariane_ost_scmp_name = "ARIANE_OS_TYPE_SUPPORTING_COMPANY_NAME"
    ariane_ost_scmp_desc = "ARIANE_OS_TYPE_SUPPORTING_COMPANY_DESCRIPTION"

    ariane_team_name = "ARIANE_TEAM_NAME"
    ariane_team_cc = "ARIANE_TEAM_COLOR_CODE"
    ariane_team_desc = "ARIANE_TEAM_DESCRIPTION"

    ariane_environment_name = "ARIANE_ENV_NAME"
    ariane_environment_cc = "ARIANE_ENV_COLOR_CODE"
    ariane_environment_desc = "ARIANE_ENV_DESCRIPTION"

    docker_props_config_cmd = "docker_cmd"
    docker_props_config_env = "docker_env"
    docker_props_config_exposed_ports = "docker_exposed_ports"
    docker_props_config_hostname = "docker_hostname"
    docker_props_config_image = "docker_image"
    docker_props_driver = "docker_driver"
    docker_props_host_config_port_binding = "docker_ports_binding"

    def __init__(self, dcontainer_id=None, mcontainer_id=None, mcontainer=None, osi_id=None, osi=None,
                 ost_id=None, ost=None, environment_id=None, environment=None, team_id=None, team=None,
                 name=None, domain=None, fqdn=None, nsenter_pid=None, details=None,
                 processs=None, last_processs=None, last_nics=None, nics=None):
        LOGGER.debug(pprint.pformat(details))
        self.did = dcontainer_id
        self.name = name
        self.domain = domain
        self.fqdn = fqdn
        self.nsented_pid = nsenter_pid
        self.details = details

        self.mid = mcontainer_id
        self.mcontainer = mcontainer
        self.oid = osi_id
        self.osi = osi
        self.ostid = ost_id
        self.ost = ost
        self.eid = environment_id
        self.environment = environment
        self.tid = team_id
        self.team = team

        self.processs = processs if processs is not None else []
        self.last_processs = last_processs if last_processs is not None else []
        self.new_processs = []

        self.last_nics = last_nics if last_nics is not None else []
        self.nics = nics if nics is not None else []

    def __eq__(self, other):
        return self.did == other.did

    def extract_properties(self):
        pass

    def extract_os_type_from_env_vars(self):
        LOGGER.debug("DockerContainer.extract_os_type_from_env_vars")
        ret = None
        if self.details is not None and self.details['Config'] and self.details['Config']['Env']:
            env_vars = self.details['Config']['Env']
            for vars in env_vars:
                if vars.startswith(DockerContainer.ariane_ost_name):
                    if ret is None:
                        ret = {}
                    ret[DockerContainer.ariane_ost_name] = vars.split('=')[1]
                if vars.startswith(DockerContainer.ariane_ost_arc):
                    if ret is None:
                        ret = {}
                    ret[DockerContainer.ariane_ost_arc] = vars.split('=')[1]
                if vars.startswith(DockerContainer.ariane_ost_scmp_name):
                    if ret is None:
                        ret = {}
                    ret[DockerContainer.ariane_ost_scmp_name] = vars.split('=')[1]
                if vars.startswith(DockerContainer.ariane_ost_scmp_desc):
                    if ret is None:
                        ret = {}
                    ret[DockerContainer.ariane_ost_scmp_desc] = vars.split('=')[1]
        return ret

    def extract_environment_from_env_vars(self):
        LOGGER.debug("DockerContainer.extract_environment_from_env_vars")
        ret = None
        if self.details is not None and self.details['Config'] and self.details['Config']['Env']:
            env_vars = self.details['Config']['Env']
            for vars in env_vars:
                if vars.startswith(DockerContainer.ariane_environment_name):
                    if ret is None:
                        ret = {}
                    ret[DockerContainer.ariane_environment_name] = vars.split('=')[1]
                if vars.startswith(DockerContainer.ariane_environment_cc):
                    if ret is None:
                        ret = {}
                    ret[DockerContainer.ariane_environment_cc] = vars.split('=')[1]
                if vars.startswith(DockerContainer.ariane_environment_desc):
                    if ret is None:
                        ret = {}
                    ret[DockerContainer.ariane_environment_desc] = vars.split('=')[1]
        return ret

    def extract_team_from_env_vars(self):
        LOGGER.debug("DockerContainer.extract_team_from_env_vars")
        ret = None
        if self.details is not None and self.details['Config'] and self.details['Config']['Env']:
            env_vars = self.details['Config']['Env']
            for vars in env_vars:
                if vars.startswith(DockerContainer.ariane_team_name):
                    if ret is None:
                        ret = {}
                    ret[DockerContainer.ariane_team_name] = vars.split('=')[1]
                if vars.startswith(DockerContainer.ariane_team_cc):
                    if ret is None:
                        ret = {}
                    ret[DockerContainer.ariane_team_cc] = vars.split('=')[1]
                if vars.startswith(DockerContainer.ariane_team_desc):
                    if ret is None:
                        ret = {}
                    ret[DockerContainer.ariane_team_desc] = vars.split('=')[1]
        return ret

    def to_json(self):
        #LOGGER.debug("DockerContainer.to_json")
        processs_2_json = []
        for process in self.processs:
            processs_2_json.append(process.to_json())
        last_processs_2_json = []
        for last_process in self.last_processs:
            last_processs_2_json.append(last_process.to_json())
        nics_2_json = []
        for nic in self.nics:
            nics_2_json.append(nic.to_json())
        last_nics_2_json = []
        for last_nic in self.last_nics:
            last_nics_2_json.append(last_nic.to_json())
        json_obj = {
            'did': self.did,
            'name': self.name,
            'domain': self.domain,
            'fqdn': self.fqdn,
            'details': self.details,
            'nsenter_pid': self.nsented_pid,
            'mid': self.mid,
            'oid': self.oid,
            'ostid': self.ostid,
            'eid': self.eid,
            'tid': self.tid,
            'processs': processs_2_json,
            'last_processs': last_processs_2_json,
            'nics': nics_2_json,
            'last_nics': last_nics_2_json
        }
        return json_obj

    @staticmethod
    def from_json(json_obj):
        #LOGGER.debug("DockerContainer.from_json")
        processs_json = json_obj['processs'] if json_obj['processs'] else []
        processs = []
        for process_json in processs_json:
            processs.append(DockerContainerProcess.from_json(process_json))
        last_processs_json = json_obj['last_processs'] if json_obj['last_processs'] else []
        last_processs = []
        for last_process_json in last_processs_json:
            last_processs.append(DockerContainerProcess.from_json(last_process_json))
        nics_json = json_obj['nics'] if json_obj['nics'] else []
        nics = []
        for nic_json in nics_json:
            nics.append(NetworkInterfaceCard.from_json(nic_json))
        last_nics_json = json_obj['last_nics'] if json_obj['last_nics'] else []
        last_nics = []
        for last_nic_json in last_nics_json:
            last_nics.append(NetworkInterfaceCard.from_json(last_nic_json))
        return DockerContainer(
            dcontainer_id=json_obj['did'] if json_obj['did'] else None,
            mcontainer_id=json_obj['mid'] if json_obj['mid'] else None,
            osi_id=json_obj['oid'] if json_obj['oid'] else None,
            ost_id=json_obj['ostid'] if json_obj['ostid'] else None,
            environment_id=json_obj['eid'] if json_obj['eid'] else None,
            team_id=json_obj['tid'] if json_obj['tid'] else None,
            name=json_obj['name'] if json_obj['name'] else None,
            domain=json_obj['domain'] if json_obj['domain'] else None,
            fqdn=json_obj['fqdn'] if json_obj['fqdn'] else None,
            details=json_obj['details'] if json_obj['details'] else None,
            nsenter_pid=json_obj['nsenter_pid'] if json_obj['nsenter_pid'] else None,
            processs=processs,
            last_processs=last_processs,
            nics=nics,
            last_nics=last_nics
        )

    def ethtool(self, iptable_nic):
        LOGGER.debug("DockerContainer.ethtool")
        if os.getuid() != 0:
            LOGGER.warning("You need to have root privileges to sniff containers namespace.")
        elif iptable_nic is None or 'name' not in iptable_nic or iptable_nic['name'] is None:
            LOGGER.error("nic name not provided on ethtool input.")
        else:
            if _platform == "linux" or _platform == "linux2":
                with Namespace(self.nsented_pid, 'net'):
                    bytes = subprocess.check_output(['ethtool', iptable_nic['name']])
                tmpfilename = tempfile.gettempdir() + os.sep + self.did + '.tmp'
                with open(tmpfilename, 'wb') as tmpfile:
                    tmpfile.write(bytes)
                    tmpfile.close()
                with open(tmpfilename, 'r') as tmpfile:
                    text = tmpfile.readlines()
                    tmpfile.close()
                os.remove(tmpfilename)
            for line in text:
                if "Duplex" in line:
                    iptable_nic['duplex'] = line.split('Duplex: ')[1].replace('\n', '')
                elif "Speed" in line:
                    iptable_nic['speed'] = line.split('Speed: ')[1].split('Mb/s')[0]
        return iptable_nic

    def ipaddr(self):
        LOGGER.debug("DockerContainer.ipaddr")
        ret = []
        if os.getuid() != 0:
            LOGGER.warning("You need to have root privileges to sniff containers namespace.")
        else:
            text = None
            if _platform == "linux" or _platform == "linux2":
                with Namespace(self.nsented_pid, 'net'):
                    bytes = subprocess.check_output(['ip', 'addr', 'show'])
                tmpfilename = tempfile.gettempdir() + os.sep + self.did + '.tmp'
                with open(tmpfilename, 'wb') as tmpfile:
                    tmpfile.write(bytes)
                    tmpfile.close()
                with open(tmpfilename, 'r') as tmpfile:
                    text = tmpfile.readlines()
                    tmpfile.close()
                os.remove(tmpfilename)

            current_card = None
            if text is not None:
                for line in text:
                    if ": <" in line and ("UP" in line or "UNKNOWN" in line):
                        card_name = line.split(":")[1].split(": <")[0].replace(' ', '')
                        if '@' in card_name:
                            card_name = card_name.split('@')[0]
                        if current_card is None:
                            current_card = {'name': card_name}
                        else:
                            ret.append(current_card)
                            current_card = {'name': card_name}
                        if "MULTICAST" in line:
                            current_card['multicast'] = True
                        else:
                            current_card['multicast'] = False
                        if "mtu" in line:
                            current_card['mtu'] = line.split("mtu")[1].split(" ")[1]

                    if "inet " in line and current_card is not None:
                        current_card['ipv4_addr'] = line.split("inet ")[1].split("/")[0]
                        z_bytes = 32 - int(line.split("/")[1].split(" ")[0])
                        if z_bytes < 8:
                            mask = "255.255.255." + str(256 - (1 << z_bytes))
                        elif z_bytes < 16:
                            mask = "255.255." + str(256 - int(1 << z_bytes - 8)) + ".0"
                        elif z_bytes < 24:
                            mask = "255." + str(256 - (1 << z_bytes - 16)) + ".0.0"
                        else:
                            mask = str(256 - (1 << z_bytes - 24)) + ".0.0.0"
                        current_card['ipv4_mask'] = mask

                    if "link" in line and "brd" in line:
                        current_card['mac_addr'] = line.split("brd")[0].split("link")[1].split(" ")[1]
                ret.append(current_card)
            return ret

    def netstat(self):
        LOGGER.debug("DockerContainer.netstat")
        ret = []
        if os.geteuid() != 0:
            LOGGER.warning("You need to have root privileges to sniff containers namespace.")
        else:
            if _platform == "linux" or _platform == "linux2":
                with Namespace(self.nsented_pid, 'net'):
                    bytes = subprocess.check_output(['netstat', '-a', '-p', '-n'])
                tmpfilename = tempfile.gettempdir() + os.sep + self.did + '.tmp'
                with open(tmpfilename, 'wb') as tmpfile:
                    tmpfile.write(bytes)
                    tmpfile.close()
                with open(tmpfilename, 'r') as tmpfile:
                    text = tmpfile.readlines()
                    tmpfile.close()
                os.remove(tmpfilename)

                for line in text:
                    if line.startswith('tcp') or line.startswith('tcp6') or \
                       line.startswith('udp') or line.startswith('udp6'):
                        fields = line.strip().split()
                        protocol = fields[0]
                        source_ep = fields[3]
                        if protocol is 'tcp' or protocol is 'udp':
                            source_ip = source_ep.split(':')[0]
                            source_port = source_ep.split(':')[1]
                        else:
                            split = source_ep.split(':')
                            split_array_length = split.__len__()
                            source_port = split[split_array_length-1]
                            source_ip = source_ep.split(':'+source_port)[0]

                        if fields.__len__() > 6:
                            state = fields[5]
                            pid = fields[6].split('/')[0]
                        else:
                            state = 'NONE'
                            pid = fields[5].split('/')[0]

                        if state != 'LISTEN' and state != 'CLOSE' and state != 'NONE':
                            target_ep = fields[4]
                            if protocol is 'tcp' or protocol is 'udp':
                                target_ip = target_ep.split(':')[0]
                                target_port = target_ep.split(':')[1]
                            else:
                                split = target_ep.split(':')
                                split_array_length = split.__len__()
                                target_port = split[split_array_length-1]
                                target_ip = target_ep.split(':'+target_port)[0]
                        else:
                            target_ip = None
                            target_port = None

                        if protocol == 'tcp' or protocol == 'udp':
                            family = 'AF_INET'
                        else:
                            family = 'AF_INET6'

                        if protocol == 'tcp' or protocol == 'tcp6':
                            type = 'SOCK_STREAM'
                        else:
                            type = 'SOCK_DGRAM'

                        ret.append({
                            'pid': pid,
                            'socket': MapSocket(source_ip=source_ip, source_port=source_port,
                                                destination_ip=target_ip, destination_port=target_port,
                                                status=state, family=family, rtype=type)
                        })
            else:
                LOGGER.warning("Containers namespace sniff enabled on Linux only.")
        return ret

    def is_in_container_destination(self, mapping_socket):
        destination_in_container = False
        if mapping_socket.family == "AF_INET6":
            for nic in self.nics:
                if nic.ipv6_address == mapping_socket.destination_ip:
                    destination_in_container = True
                    break
        if not destination_in_container:
            for nic in self.nics:
                if nic.ipv4_address == mapping_socket.destination_ip:
                    destination_in_container = True
                    break
        return destination_in_container

    def is_local_destination(self, mapping_socket):
        destination_is_local = False
        if mapping_socket.destination_ip is not None and mapping_socket.family == "AF_INET":
            for nic in self.nics:
                if NetworkInterfaceCard.ip_is_in_subnet(mapping_socket.destination_ip,
                                                        nic.ipv4_subnet_addr, nic.ipv4_subnet_mask):
                    destination_is_local = True
                    break
        elif mapping_socket.destination_ip is not None and mapping_socket.family == "AF_INET6":
            destination_ipv4 = MapSocket.ipv6_2_ipv4(mapping_socket.destination_ip)
            if destination_ipv4 != mapping_socket.destination_ip:
                for nic in self.nics:
                    if NetworkInterfaceCard.ip_is_in_subnet(destination_ipv4,
                                                            nic.ipv4_subnet_addr, nic.ipv4_subnet_mask):
                        destination_is_local = True
                        break
            else:
                #TODO: check is ipv6 in subnet ?
                for nic in self.nics:
                    if nic.ipv6_address is not None and mapping_socket.destination_ip == nic.ipv6_address:
                        destination_is_local = True
                        break
                    elif NetworkInterfaceCard.ip_is_in_subnet(mapping_socket.destination_ip,
                                                              nic.ipv4_subnet_addr, nic.ipv4_subnet_mask):
                        destination_is_local = True
                        break
        elif mapping_socket.family == "AF_UNIX":
            destination_is_local = True

        return destination_is_local

    def update(self, cli):
        LOGGER.debug("DockerContainer.update")
        self.last_processs = copy.deepcopy(self.processs)
        self.last_nics = copy.deepcopy(self.nics)
        self.sniff(cli)

    def sniff(self, cli):
        LOGGER.debug("DockerContainer.sniff")
        self.processs = []
        self.new_processs = []
        self.nics = []

        c_netstat = self.netstat()
        c_ipnics = []
        for ipaddr in self.ipaddr():
            LOGGER.debug(str(ipaddr))
            c_ipnics.append(self.ethtool(ipaddr))
        c_top = cli.top(self.did)
        top_processes = c_top['Processes'] if 'Processes' in c_top and c_top['Processes'] is not None else []

        if top_processes is None or top_processes.__len__() == 0:
            #TODO: seems to be a docker python cli bug here
            top_processes = []
            LOGGER.warning("No process found for container [" + self.did + "]" + self.name + " !?")
            LOGGER.warning(pprint.pformat(c_top))

        for processTop in top_processes:
            a_process = DockerContainerProcess(pid=processTop[1])
            if a_process in self.last_processs:
                for last_process in self.last_processs:
                    if last_process == a_process:
                        if last_process.mdpid is not None:
                            a_process.mdpid = last_process.mdpid
                        if last_process.mospid is not None:
                            a_process.mospid = last_process.mospid
                        if last_process.mcid is not None:
                            a_process.mcid = last_process.mcid
                        a_process.last_map_sockets = copy.deepcopy(last_process.map_sockets)
            else:
                self.new_processs.append(a_process)
            self.processs.append(a_process)

            for pid_socket in c_netstat:
                if pid_socket['pid'] == a_process.pid:
                    proc_sock = pid_socket['socket']
                    a_process.map_sockets.append(proc_sock)
                    if proc_sock not in a_process.last_map_sockets:
                        a_process.new_map_sockets.append(proc_sock)

            for map_socket in a_process.last_map_sockets:
                if map_socket not in a_process.new_map_sockets and \
                   map_socket.destination_node_id is None and \
                   map_socket.destination_ip is not None and map_socket.destination_port is not None:
                    LOGGER.debug('Will complete following map_socket this round : ' + map_socket.source_ip + ":" +
                                 str(map_socket.source_port))
                    a_process.new_map_sockets.append(map_socket)

        for ip_nic in c_ipnics:
            if 'ipv4_addr' in ip_nic and 'ipv4_mask' in ip_nic:
                ipv4_snet_address = \
                    str(IPv4Network(ip_nic['ipv4_addr'] + '/' + ip_nic['ipv4_mask'], strict=False).network_address)
            if self.fqdn is None:
                if 'Config' in self.details:
                    self.fqdn = self.details['Config']['Hostname'] + '.' + self.domain
                else:
                    self.fqdn = self.name + '.' + self.domain

            nic = NetworkInterfaceCard(
                name=ip_nic['name'] if 'name' in ip_nic else '',
                ipv4_address=ip_nic['ipv4_addr'] if 'ipv4_addr' in ip_nic else '',
                ipv4_fqdn=ip_nic['name'] + '.' + self.fqdn if 'name' in ip_nic else self.fqdn,
                ipv4_subnet_addr=ipv4_snet_address if ipv4_snet_address is not None else '',
                ipv4_subnet_mask=ip_nic['ipv4_mask'] if 'ipv4_mask' in ip_nic else '',
                mtu=ip_nic['mtu'] if 'mtu' in ip_nic else '',
                mac_address=ip_nic['mac_addr'] if 'mac_addr' in ip_nic else '',
                duplex=ip_nic['duplex'].upper() if 'duplex' in ip_nic else '',
                speed=ip_nic['speed'] if 'speed' in ip_nic else ''
            )
            self.nics.append(nic)


class DockerHost(object):
    def __init__(self, host_container_id=None, host_osi_id=None, host_lra_id=None,
                 hostname=None, info=None, containers=None, last_containers=None,
                 networks=None, last_networks=None):
        self.hostname = hostname
        self.info = info

        self.host_container_id = host_container_id
        self.osi_id = host_osi_id
        self.lra_id = host_lra_id

        self.containers = containers if containers is not None else []
        self.last_containers = last_containers if last_containers is not None else []
        self.new_containers = []

        self.networks = networks if networks is not None else []
        self.last_networks = last_networks if last_networks is not None else []
        self.new_networks = []

    def __eq__(self, other):
        return self.hostname == other.hostname

    def __str__(self):
        return 'docker@' + self.hostname

    def need_directories_refresh(self):
        pass

    def to_json(self):
        #LOGGER.debug("DockerHost.to_json")
        containers_2_json = []
        for container in self.containers:
            containers_2_json.append(container.to_json())
        last_containers_2_json = []
        for last_container in self.last_containers:
            last_containers_2_json.append(last_container.to_json())
        networks_2_json = []
        for network in self.networks:
            networks_2_json.append(network.to_json())
        last_networks_2_json = []
        for last_network in self.last_networks:
            last_networks_2_json.append(last_network.to_json())
        json_obj = {
            'host_container_id': self.host_container_id,
            'host_osi_id': self.osi_id,
            'host_lra_id': self.lra_id,
            'hostname': self.hostname,
            'info': self.info,
            'osi_id': self.osi_id,
            'lra_id': self.lra_id,
            'containers': containers_2_json,
            'last_containers': last_containers_2_json,
            'networks': networks_2_json,
            'last_networks': last_networks_2_json,
        }
        return json_obj

    @staticmethod
    def from_json(json_obj):
        #LOGGER.debug("DockerHost.from_json")
        containers_json = json_obj['containers'] if json_obj['containers'] else []
        containers = []
        for container_json in containers_json:
            containers.append(DockerContainer.from_json(container_json))
        last_containers_json = json_obj['last_containers'] if json_obj['last_containers'] else []
        last_containers = []
        for last_container_json in last_containers_json:
            last_containers.append(DockerContainer.from_json(last_container_json))

        networks_json = json_obj['networks'] if json_obj['networks'] else []
        networks = []
        for network_json in networks_json:
            networks.append(DockerNetwork.from_json(network_json))
        last_networks_json = json_obj['last_networks'] if json_obj['last_networks'] else []
        last_networks = []
        for last_network_json in last_networks_json:
            last_networks.append(DockerNetwork.from_json(last_network_json))

        return DockerHost(
            host_container_id=json_obj['host_container_id'] if 'host_container_id' in json_obj and json_obj['host_container_id'] else None,
            host_osi_id=json_obj['host_osi_id'] if 'host_osi_id' in json_obj and json_obj['host_osi_id'] else None,
            host_lra_id=json_obj['host_lra_id'] if 'host_lra_id' in json_obj and json_obj['host_lra_id'] else None,
            hostname=json_obj['hostname'] if 'hostname' in json_obj and json_obj['hostname'] else None,
            info=json_obj['info'] if 'info' in json_obj and json_obj['info'] else None,
            containers=containers,
            last_containers=last_containers,
            networks=networks,
            last_networks=last_networks
        )

    def update(self, cli):
        LOGGER.debug("DockerHost.update")
        self.last_containers = copy.deepcopy(self.containers)
        self.last_networks = copy.deepcopy(self.networks)
        self.sniff(cli)

    def sniff(self, cli):
        LOGGER.debug("DockerHost.sniff")
        self.containers = []
        self.networks = []
        self.new_containers = []
        self.new_networks = []

        if self.info is None:
            self.info = cli.info()
            self.hostname = self.info['Name']

        for network in cli.networks():
            bridge_name = \
                network['Options']['com.docker.network.bridge.name'] if 'com.docker.network.bridge.name' in network['Options'] is not None else None

            docker_network = DockerNetwork(
                bridge_name=bridge_name,
                nid=network['Id'],
                driver=network['Driver'],
                name=network['Name'],
                IPAM=network['IPAM'],
                options=network['Options'],
                scope=network['Scope'],
                containers_network=network['Containers']
            )
            if docker_network in self.last_networks:
                for last_network in self.last_networks:
                    if last_network == docker_network:
                        if last_network.subnet_id is not None:
                            docker_network.subnet_id = last_network.subnet_id
                        if last_network.nic_id is not None:
                            docker_network.nic_id = last_network.nic_id
            else:
                self.new_networks.append(docker_network)
            self.networks.append(docker_network)

        for container_dict in cli.containers():
            c_did = container_dict['Id']
            c_inspect = cli.inspect_container(c_did)
            c_name = c_inspect['Name'].split('/')[1]
            c_nsenterpid = c_inspect['State']['Pid']

            docker_container = DockerContainer(dcontainer_id=c_did, name=c_name, domain=self.hostname,
                                               nsenter_pid=c_nsenterpid, details=c_inspect)

            if docker_container in self.last_containers:
                for last_container in self.last_containers:
                    if last_container == docker_container:
                        if last_container.mid is not None:
                            docker_container.mid = last_container.mid
                        if last_container.eid is not None:
                            docker_container.eid = last_container.eid
                        if last_container.tid is not None:
                            docker_container.tid = last_container.tid
                        if last_container.oid is not None:
                            docker_container.oid = last_container.oid
                        else:
                            name = docker_container.name
                            LOGGER.debug('container not saved on DB on previous round: ' + name)
                            self.new_containers.append(docker_container)
                        docker_container.processs = copy.deepcopy(last_container.processs)
                        break

            else:
                self.new_containers.append(docker_container)

            self.containers.append(docker_container)
            docker_container.update(cli)


# Ariane Docker plugin
# Docker config
#
# Copyright (C) 2016 echinopsii
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import configparser
import json
import os
from ariane_docker import exceptions

__author__ = 'mffrench'


class Config(object):
    def __init__(self):
        self.rest_base_url = None
        self.rest_user = None
        self.rest_password = None

        self.rbmq_host = None
        self.rbmq_port = None
        self.rbmq_user = None
        self.rbmq_password = None
        self.rbmq_vhost = None

        self.sleeping_period = 20
        self.log_conf_file_path = "/etc/ariane/adocker_logging.json"

        self.docker_client_url = 'unix://var/run/docker.sock'
        self.docker_client_version = 'auto'
        self.docker_client_timeout = 0
        self.docker_client_tls = False

        #List of possible locations this OS instance could be located with routing area and subnets
        #(labtop of VM which can move through an hypervisor)

    def parse(self, config_file):
        if not os.path.isfile(config_file):
            raise exceptions.ArianeDockerConfigFileError(config_file)

        config_file = open(config_file, 'r')
        config = json.load(config_file)

        ariane_server_missing_fields = []
        if 'ariane_server' in config:
            self.rest_base_url = config['ariane_server']['rest_base_url']
            if self.rest_base_url is None or not self.rest_base_url:
                ariane_server_missing_fields.append('rest_base_url')

            self.rest_user = config['ariane_server']['rest_user']
            if self.rest_user is None or not self.rest_user:
                ariane_server_missing_fields.append('rest_user')

            self.rest_password = config['ariane_server']['rest_password']
            if self.rest_password is None or not self.rest_password:
                ariane_server_missing_fields.append('rest_password')

            self.rbmq_host = config['ariane_server']['rbmq_host']
            if self.rbmq_host is None or not self.rbmq_host:
                ariane_server_missing_fields.append('rbmq_host')

            self.rbmq_port = config['ariane_server']['rbmq_port']
            if self.rbmq_port is None or not self.rbmq_port:
                ariane_server_missing_fields.append('rbmq_port')

            self.rbmq_user = config['ariane_server']['rbmq_user']
            if self.rbmq_user is None or not self.rbmq_user:
                ariane_server_missing_fields.append('rbmq_user')

            self.rbmq_password = config['ariane_server']['rbmq_password']
            if self.rbmq_password is None or not self.rest_password:
                ariane_server_missing_fields.append('rbmq_password')

            self.rbmq_vhost = config['ariane_server']['rbmq_vhost']
            if self.rbmq_vhost is None or not self.rbmq_vhost:
                ariane_server_missing_fields.append('rbmq_vhost')
        else:
            raise exceptions.ArianeDockerConfigMandatorySectionMissingError('ariane_server')

        if 'ariane_docker' in config:
            if 'sleeping_period' in config['ariane_docker'] and config['ariane_docker']['sleeping_period']:
                self.sleeping_period = config['ariane_docker']['sleeping_period']
            if 'client_url' in config['ariane_docker'] and config['ariane_docker']['client_url']:
                self.docker_client_url = config['ariane_docker']['client_url']
            if 'log_conf_file_path' in config['ariane_docker'] and config['ariane_docker']['log_conf_file_path']:
                self.log_conf_file_path = config['ariane_docker']['log_conf_file_path']

        if ariane_server_missing_fields.__len__() > 0:
            raise exceptions.ArianeDockerConfigMandatoryFieldsMissingError(ariane_server_missing_fields)

        return self



# Ariane Docker plugin
# Docker component
#
# Copyright (C) 2016 echinopsii
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import json
import logging
#import pprint
import socket
import traceback
from ariane_clip3.injector import InjectorComponentSkeleton, InjectorCachedComponent
import datetime
from ariane_docker.docker import DockerHost

__author__ = 'mffrench'


LOGGER = logging.getLogger(__name__)

class DockerComponent(InjectorComponentSkeleton):

    def __init__(self, attached_gear_id=None, hostname=socket.gethostname(),
                 docker_cli=None, docker_gear_actor_ref=None):
        self.hostname = hostname
        self.docker_gear_actor_ref = docker_gear_actor_ref
        self.cli = docker_cli
        super(DockerComponent, self).__init__(
            component_id=
            'ariane.community.plugin.docker.components.cache.docker_component@' + self.hostname,
            component_name='docker_component@' + self.hostname,
            component_type="Docker injector",
            component_admin_queue=
            'ariane.community.plugin.docker.components.cache.docker_component@' + self.hostname,
            refreshing=False, next_action=InjectorCachedComponent.action_create,
            json_last_refresh=datetime.datetime.now(),
            attached_gear_id=attached_gear_id
        )
        cached_blob = self.component_cache_actor.blob.get()
        if cached_blob is not None and cached_blob:
            #LOGGER.debug("------------------------------------------------------------------")
            #LOGGER.debug("Cached blob is :\n" + pprint.pformat(cached_blob))
            #LOGGER.debug("------------------------------------------------------------------")
            self.docker_host = DockerHost.from_json(cached_blob)
        else:
            self.docker_host = DockerHost()
            self.docker_host.sniff(self.cli)
        self.version = 0

    def data_blob(self):
        data_blob = self.docker_host.to_json()
        #LOGGER.debug("------------------------------------------------------------------")
        #LOGGER.debug("Cached blob is :\n" + pprint.pformat(data_blob))
        #LOGGER.debug("------------------------------------------------------------------")
        return json.dumps(data_blob)

    def sniff(self, synchronize_with_ariane_dbs=True):
        try:
            LOGGER.info("Sniffing...")
            self.cache(refreshing=True, next_action=InjectorCachedComponent.action_update, data_blob=self.data_blob())
            self.docker_host.update(self.cli)
            self.cache(refreshing=False, next_action=InjectorCachedComponent.action_update, data_blob=self.data_blob())
            self.version += 1
            if synchronize_with_ariane_dbs and self.docker_gear_actor_ref is not None:
                self.docker_gear_actor_ref.proxy().synchronize_with_ariane_dbs()
        except Exception as e:
            LOGGER.error(e.__str__())
            LOGGER.error(traceback.format_exc())



# Ariane Docker plugin
# init
#
# Copyright (C) 2016 echinopsii
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
__author__ = 'mffrench'
__version__ = '0.1.0'

import logging
from logging import NullHandler

# Add NullHandler to prevent logging warnings
logging.getLogger(__name__).addHandler(NullHandler())


# Ariane Docker plugin
# Connectors to Ariane server and Docker Daemon
#
# Copyright (C) 2015 echinopsii
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import socket
import traceback
from docker import Client
from ariane_clip3.injector import InjectorService, InjectorUITreeEntity, InjectorUITreeService, \
    InjectorCachedComponentService, InjectorCachedGearService
from ariane_clip3.mapping import MappingService, ContainerService
from ariane_clip3.directory import DirectoryService, LocationService

__author__ = 'mffrench'

LOGGER = logging.getLogger(__name__)


class DockerConnector(object):

    def __init__(self, docker_config):
        self.ready = False
        self.cli = Client(base_url=docker_config.docker_client_url)
        no_error = True
        try:
            self.cli.info()
        except Exception as e:
            LOGGER.error("Problem while initializing docker connector on following url : " +
                         docker_config.docker_client_url)
            LOGGER.error(e.__str__())
            no_error = False

        if no_error:
            self.ready = True


class ArianeConnector(object):

    def __init__(self, docker_config):
        self.ready = False
        rest_args = {
            'type': 'REST',
            'base_url': docker_config.rest_base_url,
            'user': docker_config.rest_user,
            'password': docker_config.rest_password
        }
        client_properties = {
            'product': 'Ariane Plugin ProcOS',
            'information': 'Ariane Plugin ProcOS - Map your Operating System Process interaction and more ...',
            'ariane.pgurl': 'ssh://' + socket.gethostname(),
            'ariane.osi': socket.gethostname(),
            'ariane.otm': 'AROps',
            'ariane.app': 'Ariane',
            'ariane.cmp': 'echinopsii'
        }
        rbmq_args = {
            'type': 'RBMQ',
            'user': docker_config.rbmq_user,
            'password': docker_config.rbmq_password,
            'host': docker_config.rbmq_host,
            'port': docker_config.rbmq_port,
            'vhost': docker_config.rbmq_vhost,
            'client_properties': client_properties
        }
        self.gears_registry_cache_id = 'ariane.community.plugin.docker.gears.cache'
        docker_gears_registry_conf = {
            'registry.name': 'Ariane Docker plugin gears registry',
            'registry.cache.id': self.gears_registry_cache_id,
            'registry.cache.name': 'Ariane Docker plugin gears cache',
            'cache.mgr.name': 'ARIANE_PLUGIN_DOCKER_GEARS_CACHE_MGR'
        }
        self.components_registry_cache_id = 'ariane.community.plugin.docker.components.cache'
        docker_components_registry_conf = {
            'registry.name': 'Ariane Docker plugin components registry',
            'registry.cache.id': self.components_registry_cache_id,
            'registry.cache.name': 'Ariane Docker plugin components cache',
            'cache.mgr.name': 'ARIANE_PLUGIN_DOCKER_COMPONENTS_CACHE_MGR'
        }

        no_error = True
        DirectoryService(rest_args)
        # Test Directory Service
        try:
            LocationService.get_locations()
        except Exception as e:
            LOGGER.error("Problem while initializing Ariane directory service.")
            LOGGER.error(e.__str__())
            no_error = False

        if no_error:
            MappingService(rest_args)
            # Test Mapping Service
            try:
                ContainerService.get_containers()
            except Exception as e:
                LOGGER.error("Problem while initializing Ariane mapping service.")
                LOGGER.error(e.__str__())
                no_error = False

        if no_error:
            try:
                self.injector_service = InjectorService(
                    driver_args=rbmq_args, gears_registry_args=docker_gears_registry_conf,
                    components_registry_args=docker_components_registry_conf
                )
            except Exception as e:
                LOGGER.error("Problem while initializing Ariane injector service.")
                LOGGER.error(e.__str__())
                no_error = False

        if no_error:
            # Register UI entity if needed (and so test)
            self.injector_ui_mapping_entity = InjectorUITreeService.find_ui_tree_entity('mappingDir')
            if self.injector_ui_mapping_entity is None:
                self.injector_ui_mapping_entity = InjectorUITreeEntity(uitid="mappingDir", value="Mapping",
                                                                       uitype=InjectorUITreeEntity.entity_dir_type)
                self.injector_ui_mapping_entity.save()
            self.injector_ui_system_entity = InjectorUITreeEntity(uitid="systemDir", value="System",
                                                                  uitype=InjectorUITreeEntity.entity_dir_type,
                                                                  context_address="", description="",
                                                                  parent_id=self.injector_ui_mapping_entity.id,
                                                                  display_roles=["sysreviewer"],
                                                                  display_permissions=["injMapSysDocker:display"])
            self.injector_ui_system_entity.save()
            self.injector_ui_procos_entity = InjectorUITreeEntity(uitid="docker", value="Docker",
                                                                  uitype=InjectorUITreeEntity.entity_leaf_type,
                                                                  context_address=
                                                                  "/ariane/views/injectors/external.jsf?id=docker",
                                                                  description="Docker injector", icon="icon-docker-injector",
                                                                  parent_id=self.injector_ui_system_entity.id,
                                                                  display_roles=["sysadmin", "sysreviewer"],
                                                                  display_permissions=["injMapSysDocker:display"],
                                                                  other_actions_roles={"action": ["sysadmin"]},
                                                                  other_actions_perms={"action": ["injMapSysDocker:action"]},
                                                                  remote_injector_tree_entity_gears_cache_id=
                                                                  self.gears_registry_cache_id,
                                                                  remote_injector_tree_entity_components_cache_id=
                                                                  self.components_registry_cache_id)
            self.injector_ui_procos_entity.save()
            self.ready = True

    def stop(self):
        if self.ready:
            if InjectorCachedGearService.get_gears_cache_size() == 0 and \
                    InjectorCachedComponentService.get_components_cache_size() == 0:
                self.injector_ui_procos_entity.remove()
            self.injector_service.stop()
            self.ready = False


# ***** BEGIN LICENSE BLOCK *****
#
# For copyright and licensing please refer to LICENSE.
#
# ***** END LICENSE BLOCK *****
from setuptools import setup

long_description = ('Ariane Plugin Docker map your docker containers and more.'
                    'Where you can get more informations : '
                    '   + http://ariane.echinopsii.net'
                    '   + http://confluence.echinopsii.net/confluence/display/AD/Ariane+Documentation+Home'
                    '   + IRC on freenode #ariane.echinopsii')

setup(name='ariane_docker',
      version='0.1.1-b03',
      description='Ariane Plugin Docker',
      long_description=long_description,
      author='Mathilde Ffrench',
      author_email='mathilde.ffrench@echinopsii.net',
      maintainer='Mathilde Ffrench',
      maintainer_email='mathilde.ffrench@echinopsii.net',
      url='https://github.com/echinopsii/net.echinopsii.ariane.community.plugin.docker.git',
      download_url='https://github.com/echinopsii/net.echinopsii.ariane.community.plugin.docker.git/tarball/0.1.1-b03',
      packages=['ariane_docker'],
      license='AGPLv3',
      install_requires=['docker-py>=1.6.0', 'nsenter>=0.2',
                        'pykka>=1.2.1', 'ariane_clip3>=0.1.3-b01', 'ariane_procos>=0.1.1'],
      package_data={'': ['LICENSE', 'README.md']},
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: GNU Affero General Public License v3',
          'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: Implementation :: CPython',
          'Topic :: Communications',
          'Topic :: Internet',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: System :: Monitoring',
          'Topic :: System :: Networking'],
      zip_safe=True)



# Ariane Docker plugin
# Docker gears
#
# Copyright (C) 2016 echinopsii
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import pprint
import socket
import threading
import traceback
from ariane_clip3.directory import OSInstanceService, RoutingAreaService, SubnetService, NICardService, IPAddressService, \
    TeamService, Team, EnvironmentService, Environment, OSInstance, OSTypeService, CompanyService, Company, OSType, \
    IPAddress, NICard, LocationService
from ariane_clip3.injector import InjectorGearSkeleton
import time
import sys
from ariane_clip3.mapping import ContainerService, Container, NodeService, Node, EndpointService, Endpoint, Link, \
    Transport, TransportService
from ariane_procos.system import NetworkInterfaceCard
from ariane_docker.components import DockerComponent
from ariane_docker.docker import DockerContainer

__author__ = 'mffrench'

LOGGER = logging.getLogger(__name__)


class DirectoryGear(InjectorGearSkeleton):
    def __init__(self):
        LOGGER.debug("DirectoryGear.__init__")
        super(DirectoryGear, self).__init__(
            gear_id='ariane.community.plugin.docker.gears.cache.directory_gear@' + str(DockerHostGear.hostname),
            gear_name='docker_directory_gear@' + str(DockerHostGear.hostname),
            gear_description='Ariane Docker directory gear for ' + str(DockerHostGear.hostname),
            gear_admin_queue='ariane.community.plugin.docker.gears.cache.directory_gear@' +
                             str(DockerHostGear.hostname),
            running=False
        )
        self.update_count = 0

    def on_start(self):
        LOGGER.debug("DirectoryGear.on_start")
        self.running = True
        self.cache(running=self.running)

    def on_stop(self):
        LOGGER.debug("DirectoryGear.on_stop")
        if self.running:
            self.running = False
            self.cache(running=self.running)

    def gear_start(self):
        LOGGER.debug("DirectoryGear.gear_start")
        self.on_start()
        LOGGER.info('docker_directory_gear@' + str(DockerHostGear.hostname) + ' has been started.')

    def gear_stop(self):
        LOGGER.debug("DirectoryGear.gear_stop")
        if self.running:
            self.running = False
            self.cache(running=self.running)
            LOGGER.info('docker_directory_gear@' + str(DockerHostGear.hostname) + ' has been stopped.')

    @staticmethod
    def sync_docker_host_osi(cached_docker_host):
        LOGGER.debug("DirectoryGear.sync_docker_host_osi")
        if cached_docker_host.osi_id is not None:
            DockerHostGear.docker_host_osi = OSInstanceService.find_os_instance(osi_id=cached_docker_host.osi_id)
            if DockerHostGear.docker_host_osi.name != DockerHostGear.hostname:
                DockerHostGear.docker_host_osi = None
                cached_docker_host.osi_id = None

        if DockerHostGear.docker_host_osi is None:
            DockerHostGear.docker_host_osi = OSInstanceService.find_os_instance(osi_name=DockerHostGear.hostname)
            if DockerHostGear.docker_host_osi is None:
                LOGGER.error('Docker host ' + str(DockerHostGear.hostname) +
                             ' OS instance not found in Ariane directory')
                LOGGER.error('Did you run Ariane ProcOS on this host first ? Stopping ...')
                sys.exit(-1)
            cached_docker_host.osi_id = DockerHostGear.docker_host_osi.id

    @staticmethod
    def sync_docker_host_lra(cached_docker_host):
        LOGGER.debug("DirectoryGear.sync_docker_host_lra")
        if cached_docker_host.lra_id is not None:
            DockerHostGear.docker_host_lra = RoutingAreaService.find_routing_area(ra_id=cached_docker_host.lra_id)
            if DockerHostGear.docker_host_lra.name != DockerHostGear.hostname + '.local':
                DockerHostGear.docker_host_lra = None
                cached_docker_host.lra_id = None

        if DockerHostGear.docker_host_lra is None:
            DockerHostGear.docker_host_lra = RoutingAreaService.find_routing_area(
                ra_name=DockerHostGear.hostname + '.local'
            )
            if DockerHostGear.docker_host_lra is None:
                LOGGER.error('Docker host ' + str(DockerHostGear.hostname) +
                             ' local routing area not found in Ariane directory')
                LOGGER.error('Did you run Ariane ProcOS on this host first ? Stopping ...')
                sys.exit(-1)
            cached_docker_host.lra_id = DockerHostGear.docker_host_lra.id

    @staticmethod
    def sync_docker_networks(docker_host):
        LOGGER.debug("DirectoryGear.sync_docker_host_networks")
        # CREATION AND REMOVAL OF DOCKER NETWORK ARE DONE BY ARIANE PROCOS PLUGIN
        for docker_network in docker_host.networks:
            if docker_network in docker_host.new_networks and docker_network.bridge_name is not None:
                #SYNC NIC HOLDING THE SUBNET BRIDGE
                nic_name = DockerHostGear.hostname + '.' + docker_network.bridge_name
                nic = NICardService.find_niCard(nic_name=nic_name)
                if nic is not None:
                    docker_network.nic_id = nic.id
                    docker_network.nic = nic
                    #SYNC SUBNET
                    if nic.nic_ipa_id > 0:
                        ip_address = IPAddressService.find_ip_address(ipa_id=nic.nic_ipa_id)
                        subnet = SubnetService.find_subnet(sb_id=ip_address.ipa_subnet_id)
                        if subnet is not None:
                            docker_network.subnet_id = subnet.id
                            docker_network.subnet = subnet
                        else:
                            LOGGER.warning('docker subnet for nic ' + nic_name + ' not found in Ariane directories !')
                    else:
                        LOGGER.warning('No IP defined for NIC ' + nic_name + ' in Ariane directories !')
                else:
                    LOGGER.warning(nic_name + ' NIC not found in Ariane directories !')

    @staticmethod
    def sync_docker_container_team(docker_container):
        LOGGER.debug("DirectoryGear.sync_docker_container_team")
        team_from_conf = docker_container.extract_team_from_env_vars()
        if team_from_conf is not None:
            team_from_ariane = TeamService.find_team(team_name=team_from_conf[DockerContainer.ariane_team_name])
            if team_from_ariane is None:
                team_from_ariane = Team(
                    name=team_from_conf[DockerContainer.ariane_team_name],
                    color_code=team_from_conf[DockerContainer.ariane_team_cc],
                    description=team_from_conf[DockerContainer.ariane_team_desc]
                )
                try:
                    team_from_ariane.save()
                except Exception as e:
                    LOGGER.warning("Unable to save team (" + str(team_from_conf) + ") in Ariane Directories !")
                    LOGGER.warning(e.__str__())
                    LOGGER.warning(traceback.format_exc())

            if team_from_ariane is not None:
                docker_container.team = team_from_ariane
                docker_container.tid = team_from_ariane.id
        else:
            LOGGER.warning("Team is not specified in the docker container ( " + docker_container.name +
                           " ) environment variables !")

        if docker_container.team is None and DockerHostGear.docker_host_osi.team_ids.__len__()>0:
            LOGGER.warning("Docker container team will be herited from first host team !")
            docker_container.team = TeamService.find_team(team_id=DockerHostGear.docker_host_osi.team_ids[0])
            docker_container.tid = DockerHostGear.docker_host_osi.team_ids[0]

    @staticmethod
    def sync_docker_container_env(docker_container):
        LOGGER.debug("DirectoryGear.sync_docker_container_env")
        env_from_conf = docker_container.extract_environment_from_env_vars()
        if env_from_conf is not None:
            env_from_ariane = EnvironmentService.find_environment(
                env_name=env_from_conf[DockerContainer.ariane_environment_name]
            )
            if env_from_ariane is None:
                env_from_ariane = Environment(
                    name=env_from_conf[DockerContainer.ariane_environment_name],
                    color_code=env_from_conf[DockerContainer.ariane_environment_cc],
                    description=env_from_conf[DockerContainer.ariane_environment_desc]
                )
                try:
                    env_from_ariane.save()
                except Exception as e:
                    LOGGER.warning("Unable to save environment (" + str(env_from_conf) + ") in Ariane Directories !")
                    LOGGER.warning(e.__str__())
                    LOGGER.warning(traceback.format_exc())

            if env_from_ariane is not None:
                docker_container.environment = env_from_ariane
                docker_container.eid = env_from_ariane.id
        else:
            LOGGER.warning("Environment is not specified in the docker container ( " + docker_container.name +
                           " ) environment variables !")

        if docker_container.environment is None and DockerHostGear.docker_host_osi.environment_ids.__len__()>0:
            LOGGER.warning("Docker container environment will be herited from first host environment !")
            docker_container.environment = EnvironmentService.find_environment(
                env_id=DockerHostGear.docker_host_osi.environment_ids[0]
            )
            docker_container.eid = DockerHostGear.docker_host_osi.team_ids[0]

    @staticmethod
    def sync_docker_container_ost(docker_container):
        LOGGER.debug("DirectoryGear.sync_docker_container_ost")
        ost_from_conf = docker_container.extract_os_type_from_env_vars()
        if ost_from_conf is not None:
            ost_from_ariane = OSTypeService.find_ostype(ost_name=ost_from_conf[DockerContainer.ariane_ost_name])
            if ost_from_ariane is None:
                cmp_from_ariane = CompanyService.find_company(
                    cmp_name=ost_from_conf[DockerContainer.ariane_ost_scmp_name]
                )
                if cmp_from_ariane is None:
                    cmp_from_ariane = Company(
                        name=ost_from_conf[DockerContainer.ariane_ost_scmp_name],
                        description=ost_from_conf[DockerContainer.ariane_ost_scmp_desc]
                    )
                    cmp_from_ariane.save()
                ost_from_ariane = OSType(
                    name=ost_from_conf[DockerContainer.ariane_ost_name],
                    architecture=ost_from_conf[DockerContainer.ariane_ost_arc],
                    os_type_company_id=cmp_from_ariane.id
                )
                ost_from_ariane.save()
            docker_container.ostid = ost_from_ariane.id
            docker_container.ost = ost_from_ariane
        else:
            LOGGER.warning("OS Type is not specified in the docker container ( " + docker_container.name +
                           " ) environment variables !")

    @staticmethod
    def sync_docker_container_osi(docker_container):
        LOGGER.debug("DirectoryGear.sync_docker_container_osi")
        osi_from_ariane = OSInstanceService.find_os_instance(
            osi_name=docker_container.name + '.' + DockerHostGear.hostname
        )
        if osi_from_ariane is None:
            env_ids = [docker_container.eid] if docker_container.eid is not None else None
            team_ids = [docker_container.tid] if docker_container.tid is not None else None
            osi_from_ariane = OSInstance(
                name=docker_container.name + '.' + DockerHostGear.hostname,
                description=docker_container.name + '@' + DockerHostGear.hostname,
                admin_gate_uri=DockerHostGear.docker_host_osi.admin_gate_uri + '/$[docker exec -i -t ' +
                               docker_container.name + ' /bin/bash]',
                osi_embedding_osi_id=DockerHostGear.docker_host_osi.id,
                osi_ost_id=docker_container.ostid,
                osi_environment_ids=env_ids,
                osi_team_ids=team_ids
            )
            osi_from_ariane.save()
        docker_container.oid = osi_from_ariane.id
        docker_container.osi = osi_from_ariane

    @staticmethod
    def sync_docker_container_ip_and_nic(docker_container, docker_host_networks):
        LOGGER.debug("DirectoryGear.sync_docker_container_ip_and_nic")
        for docker_container_nic in docker_container.nics:
            if docker_container_nic not in docker_container.last_nics:
                if not docker_container_nic.ipv4_address.startswith('127'):
                    #LOGGER.debug(str(docker_container_nic))
                    for docker_network in docker_host_networks:
                        if docker_network.subnet is None and docker_network.subnet_id is not None:
                            docker_network.subnet = SubnetService.find_subnet(sb_name=docker_network.subnet_id)

                        if docker_network.subnet is not None:
                            dock_subnet = docker_network.subnet
                            if NetworkInterfaceCard.ip_is_in_subnet(docker_container_nic.ipv4_address,
                                                                    docker_network.subnet.ip,
                                                                    docker_network.subnet.mask):
                                ip_address = IPAddressService.find_ip_address(
                                    ipa_ip_address=docker_container_nic.ipv4_address,
                                    ipa_subnet_id=docker_network.subnet.id
                                )
                                if ip_address is None:
                                    ip_address = IPAddress(ip_address=docker_container_nic.ipv4_address,
                                                           fqdn=docker_container_nic.ipv4_fqdn,
                                                           ipa_subnet_id=docker_network.subnet.id,
                                                           ipa_osi_id=docker_container.osi.id)
                                    ip_address.save()
                                    docker_network.subnet.sync()
                                else:
                                    if ip_address.ipa_os_instance_id != docker_container.osi.id:
                                        ip_address.ipa_os_instance_id = docker_container.osi.id
                                        ip_address.save()
                                nicmcaddr = docker_container_nic.mac_address
                                #LOGGER.debug(str(ip_address))
                #else:
                #    TODO: currently docker container local subnet and routing area.
                #    TODO: need to enable multiple routing area for one subnet on ariane directories
                #    NOTE: anyway this feature will probably not needed as docker routing is done from the host
                #    loopback_subnet_conf = SubnetConfig(
                #        name=docker_container.name + "." + docker_container.domain + ".loopback",
                #        description=docker_container.name + "." + docker_container.domain + " loopback subnet",
                #        subnet_ip="127.0.0.0",
                #        subnet_mask="255.0.0.0"
                #    )
                #    pass

                                if dock_subnet is not None:
                                    docker_container.osi.add_subnet(dock_subnet)
                                    docker_container.osi.sync()

                                if nicmcaddr is not None and nicmcaddr:
                                    nic2save = NICardService.find_niCard(nic_mac_Address=nicmcaddr)
                                    if nic2save is None:
                                        nic2save = NICard(
                                            name=docker_container.domain + "." + docker_container.name + "." +
                                                 docker_container_nic.name,
                                            macAddress=nicmcaddr,
                                            duplex=docker_container_nic.duplex,
                                            speed=docker_container_nic.speed,
                                            mtu=docker_container_nic.mtu,
                                            nic_osi_id=docker_container.osi.id,
                                            nic_ipa_id=ip_address.id if ip_address is not None else None
                                        )
                                        nic2save.save()
                                    else:
                                        nic2save.nic_ipa_id = ip_address.id if ip_address is not None else None
                                        nic2save.sync()
                                    docker_container_nic.nic_id = nic2save.id

    @staticmethod
    def sync_docker_containers(docker_host):
        LOGGER.debug("DirectoryGear.sync_docker_containers")
        for docker_container in docker_host.new_containers:
            DirectoryGear.sync_docker_container_team(docker_container)
            DirectoryGear.sync_docker_container_env(docker_container)
            DirectoryGear.sync_docker_container_ost(docker_container)
            DirectoryGear.sync_docker_container_osi(docker_container)
            DirectoryGear.sync_docker_container_ip_and_nic(docker_container, docker_host.networks)
            docker_container.osi.sync()

        #LOGGER.debug("last containers: " + pprint.pformat(docker_host.last_containers))
        #LOGGER.debug("current containers: " + pprint.pformat(docker_host.containers))
        for docker_container in docker_host.last_containers:
            if docker_container not in docker_host.containers:
                if docker_container.oid is not None:
                    osi_from_ariane = OSInstanceService.find_os_instance(
                        osi_id=docker_container.oid
                    )
                else:
                    osi_from_ariane = OSInstanceService.find_os_instance(
                        osi_name=docker_container.name + '.' + DockerHostGear.hostname
                    )
                if osi_from_ariane is not None:
                    osi_from_ariane.remove()
            else:
                for nic in docker_container.last_nics:
                    if nic not in docker_container.nics:
                        nic2rm = NICardService.find_niCard(nic_id=nic.nic_id)
                        if nic2rm is not None:
                            ip2rm = IPAddressService.find_ip_address(ipa_id=nic2rm.nic_ipa_id)
                            if ip2rm is not None:
                                ip2rm.remove()
                            else:
                                LOGGER.warning("IP " + ip2rm.ip_address + " already removed !?")
                            nic2rm.remove()
                        else:
                            LOGGER.warning("NIC " + nic2rm.macAddress + " already removed !?")
                DirectoryGear.sync_docker_container_ip_and_nic(docker_container, docker_host.networks)

    def update_ariane_directories(self, docker_host):
        LOGGER.debug("DirectoryGear.update_ariane_directories")
        if docker_host.networks != docker_host.last_networks:
            self.sync_docker_networks(docker_host)
        else:
            LOGGER.debug("No changes on docker host networks !")

        if docker_host.containers != docker_host.last_containers:
            self.sync_docker_containers(docker_host)
        else:
            LOGGER.debug("No changes on docker host containers !")

    def init_ariane_directories(self, component):
        LOGGER.debug("DirectoryGear.init_ariane_directories")
        docker_host = component.docker_host.get()
        self.sync_docker_host_osi(docker_host)
        self.sync_docker_host_lra(docker_host)
        self.sync_docker_networks(docker_host)
        self.sync_docker_containers(docker_host)

    def synchronize_with_ariane_directories(self, component):
        LOGGER.debug("DirectoryGear.synchronize_with_ariane_directories")
        if self.running:
            docker_host = component.docker_host.get()
            self.update_ariane_directories(docker_host)
            self.update_count += 1
        else:
            LOGGER.warning("Synchronization requested but docker_directory_gear@" + str(DockerHostGear.hostname) +
                           " is not running.")


class MappingGear(InjectorGearSkeleton):

    docker_host_mco = None

    def __init__(self):
        LOGGER.debug("MappingGear.__init__")
        super(MappingGear, self).__init__(
            gear_id='ariane.community.plugin.docker.gears.cache.mapping_gear@' + str(DockerHostGear.hostname),
            gear_name='docker_mapping_gear@' + str(DockerHostGear.hostname),
            gear_description='Ariane Docker injector gear for ' + str(DockerHostGear.hostname),
            gear_admin_queue='ariane.community.plugin.docker.gears.cache.mapping_gear@' + str(DockerHostGear.hostname),
            running=False
        )
        self.update_count = 0

    def on_start(self):
        LOGGER.debug("MappingGear.on_start")
        self.running = True
        self.cache(running=self.running)

    def on_stop(self):
        LOGGER.debug("MappingGear.on_stop")
        if self.running:
            self.running = False
            self.cache(running=self.running)

    def gear_start(self):
        LOGGER.debug("MappingGear.gear_start")
        self.on_start()
        LOGGER.warning('docker_mapping_gear@' + str(DockerHostGear.hostname) + ' has been started.')

    def gear_stop(self):
        LOGGER.debug("MappingGear.gear_stop")
        if self.running:
            self.running = False
            self.cache(running=self.running)
            LOGGER.warning('docker_mapping_gear@' + str(DockerHostGear.hostname) + ' has been stopped.')

    @staticmethod
    def synchronize_new_map_socket(docker_container, process, map_socket):
        LOGGER.debug("MappingGear.synchronize_new_map_socket")
        if map_socket.source_ip is not None and map_socket.source_port is not None:
            proto = None
            if map_socket.type == "SOCK_STREAM":
                proto = "tcp://"
            elif map_socket.type == "SOCK_DGRAM":
                proto = "udp://"
            else:
                LOGGER.warning("socket type " + map_socket.type + " currently not supported !")

            if proto is not None:
                source_endpoint = None
                source_url = proto + map_socket.source_ip + ":" + str(map_socket.source_port)

                target_endpoint = None
                target_url = None

                if map_socket.source_endpoint_id is None:
                    if process.mdp is None:
                        parent_node = NodeService.find_node(nid=process.mdpid)
                    else:
                        parent_node = process.mdp
                    if parent_node is not None:
                        source_endpoint = Endpoint(
                            url=source_url,
                            parent_node=parent_node
                        )
                        source_endpoint.save()
                        map_socket.source_endpoint_id = source_endpoint.id
                    else:
                        LOGGER.warning("Fail to sync parent node for source endpoint " + source_url)
                else:
                    source_endpoint = EndpointService.find_endpoint(eid=map_socket.source_endpoint_id)

                if map_socket.destination_ip is not None and map_socket.destination_port is not None:
                    target_url = proto + map_socket.destination_ip + ":" + str(map_socket.destination_port)

                if target_url is not None:
                    if map_socket.destination_endpoint_id is None:
                        if docker_container.is_local_destination(map_socket):
                            if not docker_container.is_in_container_destination(map_socket):
                                selector = "endpointURL =~ '" + target_url + ".*'"
                                endpoints = EndpointService.find_endpoint(selector=selector)
                                if endpoints is not None:
                                    if endpoints.__len__() == 1:
                                        target_endpoint = endpoints[0]
                                    else:
                                        LOGGER.warning("Several endpoints found for selector " + selector +
                                                       ". There should be one endpoint only !")
                                else:
                                    LOGGER.warning("No endpoint for selector " + selector + "  ?!")
                            else:
                                target_local_process = None
                                mirror_map_socket = None
                                for dc_process in docker_container.processs:
                                    for m_map_socket in dc_process.map_sockets:
                                        if (m_map_socket.source_ip + ":" + m_map_socket.source_port) == target_url:
                                            target_local_process = dc_process
                                            mirror_map_socket = m_map_socket
                                            break
                                if target_local_process is not None and mirror_map_socket is not None:
                                    if mirror_map_socket.source_endpoint_id is None:
                                        if target_local_process.mdp is None:
                                            parent_node = NodeService.find_node(nid=target_local_process.mdpid)
                                        else:
                                            parent_node = target_local_process.mdp

                                        if parent_node is not None:
                                            target_endpoint = Endpoint(
                                                url=target_url,
                                                parent_node=parent_node
                                            )
                                            target_endpoint.save()
                                            mirror_map_socket.source_endpoint_id = target_endpoint.id
                                        else:
                                            LOGGER.warning("Fail to sync parent node for target endpoint " + target_url)
                                    else:
                                        target_endpoint = EndpointService.find_endpoint(
                                            eid=mirror_map_socket.source_endpoint_id
                                        )
                        else:
                            LOGGER.warning("Non local docker link not supported currently !")
                    else:
                        target_endpoint = EndpointService.find_endpoint(map_socket.destination_endpoint_id)

                    if source_endpoint is None:
                        LOGGER.warning("Unable to define source endpoint: " + source_url)
                    else:
                        if target_endpoint is None:
                            LOGGER.warning("Unable to define target endpoint: " + target_url)
                        else:
                            map_socket.destination_endpoint_id = target_endpoint.id
                            map_socket.destination_node_id = target_endpoint.parent_node_id

                            transport = Transport(name=proto)
                            if transport.id is None:
                                transport.save()

                            if transport.id is not None:
                                link = Link(source_endpoint_id=map_socket.source_endpoint_id,
                                            target_endpoint_id=map_socket.destination_endpoint_id,
                                            transport_id=transport.id)
                                link.save()
                                map_socket.transport_id = transport.id
                                map_socket.link_id = link.id

        else:
            LOGGER.warning('no source ip / port - ' + str(map_socket))

    @staticmethod
    def synchronize_removed_map_socket(docker_container, process, map_socket):
        LOGGER.debug("MappingGear.synchronize_removed_map_socket")
        if map_socket.source_endpoint_id is not None:
            source_endpoint = EndpointService.find_endpoint(eid=map_socket.source_endpoint_id)
            if source_endpoint is not None:
                LOGGER.debug("Remove (source) endpoint " + source_endpoint.url)
                source_endpoint.remove()
            else:
                LOGGER.warning("Dead socket (source endpoint : " + str(map_socket.source_endpoint_id) +
                               ") doesn't exist anymore on DB!")

        #if map_socket.destination_endpoint_id is not None:
        #    destination_endpoint = EndpointService.find_endpoint(eid=map_socket.destination_endpoint_id)
        #    if destination_endpoint is not None:
        #        LOGGER.debug("Remove (destination) endpoint " + destination_endpoint.url)
        #        destination_endpoint.remove()
        #    else:
        #        LOGGER.warning("Dead socket (destination endpoint : " + str(map_socket.source_endpoint_id) +
        #                       ") doesn't exist anymore on DB!")

    @staticmethod
    def synchronize_process_sockets(docker_container, process):
        LOGGER.debug("MappingGear.synchronize_process_sockets")
        LOGGER.debug("current sockets: " + pprint.pformat(process.map_sockets))
        LOGGER.debug("new sockets: " + pprint.pformat(process.new_map_sockets))
        LOGGER.debug("last sockets: " + pprint.pformat(process.last_map_sockets))
        for map_socket in process.new_map_sockets:
            MappingGear.synchronize_new_map_socket(docker_container, process, map_socket)
        for map_socket in process.last_map_sockets:
            if map_socket not in process.map_sockets:
                MappingGear.synchronize_removed_map_socket(docker_container, process, map_socket)

    @staticmethod
    def synchronize_process_properties(docker_container, process):
        LOGGER.debug("MappingGear.synchronize_processs_properties")
        if process.mosp is None and process.mospid is not None:
            process.mosp = NodeService.find_node(nid=process.mospid)
        if process.mosp is not None:
            process.mosp.sync()
            if process.mdp is None and process.mdpid is not None:
                process.mdp = NodeService.find_node(nid=process.mdpid)
            if process.mdp is not None:
                process.mdp.sync()
                if process.mosp.properties is not None:
                    for key in process.mosp.properties.keys():
                        process.mdp.add_property((key, process.mosp.properties[key]))
                else:
                    LOGGER.warning("No properties for process " +
                                   str(process.mosp.name) if process.mosp.name is not None else "???" + " ?!")
            else:
                LOGGER.warning("Process " + process.name + " has been lost on mapping DB ?!")
        else:
            LOGGER.warning("Shadow process of " + process.name + " has been lost on mapping DB ?!")

    @staticmethod
    def synchronize_new_processs_node(docker_container, process):
        LOGGER.debug("MappingGear.synchronize_new_processs_node")
        mapping_container = docker_container.mcontainer
        process.mcid = mapping_container.id

        mosp = None
        MappingGear.docker_host_mco.sync()

        nodes_to_test = NodeService.find_node(selector="nodeName =~ '.*" + str(process.pid) + ".*'")
        if nodes_to_test is not None:
            for node_to_test in nodes_to_test:
                if node_to_test.name.startswith('[' + str(process.pid) + ']'):
                    LOGGER.debug("Shadow Mapping OS node found : " + node_to_test.name)
                    mosp = node_to_test
                    break

        if mosp is not None:
            process_node = Node(
                name=mosp.name,
                container_id=process.mcid
            )
            process_node.save()
            process.mdpid = process_node.id
            process.mdp = process_node
            process.mospid = mosp.id
            process.mosp = mosp
        else:
            LOGGER.warning("Shadow Mapping OS node for process " + str(process.pid) + " not found !")

    @staticmethod
    def synchronize_removed_processs_node(docker_container, process):
        LOGGER.debug("MappingGear.synchronize_removed_processs_node")
        mapping_container = docker_container.mcontainer
        if mapping_container is None and docker_container.mid is not None:
            docker_container.mcontainer = ContainerService.find_container(cid=docker_container.mid)
            mapping_container = docker_container.mcontainer
        if mapping_container is not None:
            process_node = NodeService.find_node(nid=process.mdpid)
            if process_node is not None:
                process_node.remove()
            else:
                LOGGER.warning("Mapping node for process " + process.pid + "@" + mapping_container.name +
                               " not found !")
        else:
            LOGGER.warning("Mapping container not found for container " + docker_container.name + " !?")

    @staticmethod
    def synchronize_container_processs(docker_container):
        LOGGER.debug("MappingGear.synchronize_container_processs")
        # SYNC NEW PROCESSES NODES
        for process in docker_container.new_processs:
            MappingGear.synchronize_new_processs_node(docker_container, process)
        # SYNC CURRENT PROCESSES AND SOCKETS
        for process in docker_container.processs:
            MappingGear.synchronize_process_properties(docker_container, process)
            MappingGear.synchronize_process_sockets(docker_container, process)
        # SYNC DEAD PROCESSES
        for process in docker_container.last_processs:
            if process not in docker_container.processs:
                MappingGear.synchronize_removed_processs_node(docker_container, process)

    @staticmethod
    def synchronize_container_properties(docker_container):
        LOGGER.debug("MappingGear.synchronize_container_properties")
        mapping_container = docker_container.mcontainer
        if docker_container.osi is None and docker_container.oid is not None:
            docker_container.osi = OSInstanceService.find_os_instance(osi_id=docker_container.oid)

        if docker_container.team is None and docker_container.tid is not None:
            docker_container.team = TeamService.find_team(team_id=docker_container.tid)

        is_ok = True
        if docker_container.osi is not None:
            docker_container.osi.sync()
        else:
            is_ok = False
            LOGGER.warning("Docker container " + docker_container.name + " linked OS instance not found !?")

        if is_ok and docker_container.team is not None:
            docker_container.team.sync()
        else:
            is_ok = False
            LOGGER.warning("Docker container " + docker_container.name + " linked team not found !?")

        if is_ok:
            if DockerHostGear.docker_host_lra.loc_ids.__len__() > 1:
                LOGGER.warning("Localhost routing area " + DockerHostGear.docker_host_lra.name +
                               " have more than one location ?!")
            elif DockerHostGear.docker_host_lra.loc_ids.__len__() == 1:
                location = LocationService.find_location(DockerHostGear.docker_host_lra.loc_ids[0])
                if location is not None:
                    location_properties = {
                        Container.PL_NAME_MAPPING_FIELD: location.name,
                        Container.PL_ADDR_MAPPING_FIELD: location.address,
                        Container.PL_TOWN_MAPPING_FIELD: location.town,
                        Container.PL_CNTY_MAPPING_FIELD: location.country,
                        Container.PL_GPSA_MAPPING_FIELD: location.gpsLatitude,
                        Container.PL_GPSN_MAPPING_FIELD: location.gpsLongitude
                    }
                    mapping_container.add_property((Container.PL_MAPPING_PROPERTIES, location_properties))
                else:
                    LOGGER.warning("Location ( " + DockerHostGear.docker_host_lra.loc_ids[0] + " ) not found ?!")
            else:
                LOGGER.warning("Localhost routing area " + DockerHostGear.docker_host_lra.name +
                               " don't have location ?!")

            network_properties = []
            ra_subnets = {}
            ra_list = []
            for subnet_id in docker_container.osi.subnet_ids:
                subnet = SubnetService.find_subnet(subnet_id)
                if subnet is not None:
                    if subnet.routing_area_id is not None:
                        if subnet.routing_area_id not in ra_subnets:
                            routing_area = RoutingAreaService.find_routing_area(ra_id=subnet.routing_area_id)
                            if routing_area is not None:
                                ra_list.append(routing_area)
                                ra_subnets[subnet.routing_area_id] = []
                                ra_subnets[subnet.routing_area_id].append({
                                    Container.SUBNET_NAME_MAPPING_FIELD: subnet.name,
                                    Container.SUBNET_IPAD_MAPPING_FIELD: subnet.ip,
                                    Container.SUBNET_MASK_MAPPING_FIELD: subnet.mask,
                                    Container.SUBNET_ISDEFAULT_MAPPING_FIELD: subnet.is_default
                                })
                            else:
                                LOGGER.warning("Routing Area ( " + subnet.routing_area_id + " ) for subnet (" +
                                               subnet.name + ") not fount !?")
                        else:
                            ra_subnets[subnet.routing_area_id].append({
                                Container.SUBNET_NAME_MAPPING_FIELD: subnet.name,
                                Container.SUBNET_IPAD_MAPPING_FIELD: subnet.ip,
                                Container.SUBNET_MASK_MAPPING_FIELD: subnet.mask,
                                Container.SUBNET_ISDEFAULT_MAPPING_FIELD: subnet.is_default
                            })
                else:
                   LOGGER.warning("Subnet (" + subnet_id + ") not found !?")

            for ra in ra_list:
                network_properties.append(
                    {
                        Container.RAREA_NAME_MAPPING_FIELD: ra.name,
                        Container.RAREA_MLTC_MAPPING_FIELD: ra.multicast,
                        Container.RAREA_TYPE_MAPPING_FIELD: ra.type,
                        Container.RAREA_SUBNETS: ra_subnets[ra.id]
                    })
            if network_properties.__len__() > 0:
                mapping_container.add_property((Container.NETWORK_MAPPING_PROPERTIES, network_properties))

            team_properties = {
                    Container.TEAM_NAME_MAPPING_FIELD: docker_container.team.name,
                    Container.TEAM_COLR_MAPPING_FIELD: docker_container.team.color_code
            }
            mapping_container.add_property((Container.TEAM_SUPPORT_MAPPING_PROPERTIES, team_properties))

            mapping_container.add_property((
                DockerContainer.docker_props_config_image,
                docker_container.details['Config']['Image']
            ))

            exposed_ports = docker_container.details['Config']['ExposedPorts'].keys()
            if exposed_ports.__len__() > 0:
                mapping_container.add_property((
                    DockerContainer.docker_props_config_exposed_ports,
                    exposed_ports
                ))

            mapping_container.add_property((
                DockerContainer.docker_props_config_cmd,
                docker_container.details['Config']['Cmd']
            ))

            env = {}
            for envvar in docker_container.details['Config']['Env']:
                envvar_name = envvar.split('=')[0]
                if 'ARIANE' not in envvar_name:
                    envvar_value = None
                    if 'PASSWORD' in envvar_name or 'password' in envvar_name or 'PWD' in envvar_name or 'pwd' in envvar_name:
                        envvar_value = '*****'
                    else:
                        envvar_value = envvar.split('=')[1]
                    env[envvar_name] = envvar_value
            if env.keys().__len__() > 0:
                mapping_container.add_property((
                    DockerContainer.docker_props_config_env,
                    env
                ))

            mapping_container.add_property((
                DockerContainer.docker_props_config_hostname,
                docker_container.details['Config']['Hostname']
            ))

            mapping_container.add_property((
                DockerContainer.docker_props_driver,
                docker_container.details['Driver']
            ))

            ports_binding = []
            ports_binding_keys = docker_container.details['HostConfig']['PortBindings']
            if ports_binding_keys.__len__() > 0:
                for exposed_port in ports_binding_keys:
                    targets = docker_container.details['HostConfig']['PortBindings'][exposed_port]
                    if targets.__len__() > 0:
                        port_binding = exposed_port + ' <-> '
                        for target in targets:
                            host_ip = target['HostIp']
                            host_port = target['HostPort']
                            if host_ip == '':
                                host_ip = '0.0.0.0'
                            port_binding += host_ip + ':' + host_port + ' '
                        ports_binding.append(port_binding)
            mapping_container.add_property((
                DockerContainer.docker_props_host_config_port_binding,
                ports_binding
            ))

    @staticmethod
    def synchronize_existing_container(docker_container):
        LOGGER.debug("MappingGear.synchronize_existing_containers")
        MappingGear.synchronize_container_processs(docker_container)

    @staticmethod
    def synchronize_new_container(docker_container):
        LOGGER.debug("MappingGear.synchronize_new_container")
        if docker_container.ost is None:
            if docker_container.ostid is not None:
                docker_container.ost = OSTypeService.find_ostype(ost_id=docker_container.ostid)

        product = "Docker Container"
        company = "Docker Inc."
        if docker_container.ost is not None:
            if docker_container.ost.company is None:
                docker_container.ost.company = CompanyService.find_company(docker_container.ost.company_id)
            if docker_container.ost.company is not None:
                product = docker_container.ost.name + ' - ' + docker_container.ost.architecture
                company = docker_container.ost.company.name

        mapping_container = Container(
            name=docker_container.name,
            gate_uri=DockerHostGear.docker_host_osi.admin_gate_uri + '/$[docker exec -i -t ' +
                     docker_container.name + ' /bin/bash]',
            primary_admin_gate_name='NamespaceAccess@'+docker_container.name,
            parent_container_id=MappingGear.docker_host_mco.id,
            company=company,
            product=product,
            c_type="Docker Container"
        )
        mapping_container.save()
        LOGGER.debug(pprint.pformat(mapping_container.__dict__))
        docker_container.mid = mapping_container.id
        docker_container.mcontainer = mapping_container
        MappingGear.synchronize_container_properties(docker_container)
        MappingGear.synchronize_container_processs(docker_container)

    @staticmethod
    def synchronize_removed_container(docker_container):
        LOGGER.debug("MappingGear.synchronize_removed_container - " + docker_container.name)
        docker_container.mcontainer.remove()

    def synchronize_container(self, docker_host):
        LOGGER.debug("MappingGear.synchronize_container")
        LOGGER.debug("current containers: " + pprint.pformat(docker_host.containers))
        LOGGER.debug("new containers: " + pprint.pformat(docker_host.new_containers))
        LOGGER.debug("last containers: " + pprint.pformat(docker_host.last_containers))
        for docker_container in docker_host.last_containers:
            if docker_container.mcontainer is None:
                if docker_container.mid is None:
                    docker_container.mcontainer = ContainerService.find_container(
                        primary_admin_gate_url=DockerHostGear.docker_host_osi.admin_gate_uri + '/$[docker exec -i -t ' +
                                               docker_container.name + ' /bin/bash]'
                    )
                    if docker_container.mcontainer is not None:
                        docker_container.mid = docker_container.mcontainer.id
                else:
                    docker_container.mcontainer = ContainerService.find_container(cid=docker_container.mid)
            if docker_container.mcontainer is not None:
                self.synchronize_existing_container(docker_container)
            else:
                LOGGER.warning("Mapping container not found for container to be updated (" +
                               docker_container.name + ") !?")

        for docker_container in docker_host.new_containers:
            self.synchronize_new_container(docker_container)

        for docker_container in docker_host.last_containers:
            if docker_container not in docker_host.containers:
                if docker_container.mcontainer is None:
                    if docker_container.mid is None:
                        docker_container.mcontainer = ContainerService.find_container(
                            primary_admin_gate_url=DockerHostGear.docker_host_osi.admin_gate_uri + '/$[docker exec -i -t ' +
                                                   docker_container.name + ' /bin/bash]'
                        )
                        if docker_container.mcontainer is not None:
                            docker_container.mid = docker_container.mcontainer.id
                    else:
                        docker_container.mcontainer = ContainerService.find_container(cid=docker_container.mid)
                if docker_container.mcontainer is not None:
                    self.synchronize_removed_container(docker_container)
                else:
                    LOGGER.warning("Mapping container not found for container to be removed (" +
                                   docker_container.name + ") !?")

    def init_ariane_mapping(self, component):
        LOGGER.debug("MappingGear.init_ariane_mapping")
        MappingGear.docker_host_mco = ContainerService.find_container(
            primary_admin_gate_url=DockerHostGear.docker_host_osi.admin_gate_uri
        )
        if MappingGear.docker_host_mco is None:
            LOGGER.error('Docker host ' + str(DockerHostGear.hostname) +
                         ' Ariane container not found in Ariane mapping DB')
            LOGGER.error('Did you run Ariane ProcOS on this host first ? Stopping ...')
            sys.exit(-1)
        else:
            docker_host = component.docker_host.get()
            try:
                self.synchronize_container(docker_host)
            except Exception as e:
                LOGGER.error(e.__str__())
                LOGGER.error(traceback.format_exc())

    def synchronize_with_ariane_mapping(self, component):
        LOGGER.debug("MappingGear.synchronize_with_ariane_mapping")
        if self.running:
            docker_host = component.docker_host.get()
            try:
                self.synchronize_container(docker_host)
            except Exception as e:
                LOGGER.error(e.__str__())
                LOGGER.error(traceback.format_exc())
            self.update_count += 1
        else:
            LOGGER.warning('Synchronization requested but docker_mapping_gear@' +
                           str(DockerHostGear.hostname) + ' is not running.')


class DockerHostGear(InjectorGearSkeleton):
    #static reference on commons var
    config = None
    hostname = None

    docker_host_osi = None
    docker_host_lra = None

    def __init__(self, config, cli):
        LOGGER.debug("DockerHostGear.__init__")
        DockerHostGear.hostname = socket.gethostname()
        DockerHostGear.config = config
        super(DockerHostGear, self).__init__(
            gear_id='ariane.community.plugin.docker.gears.cache.docker_host_gear@' + str(DockerHostGear.hostname),
            gear_name='docker_host_gear@' + str(DockerHostGear.hostname),
            gear_description='Ariane Docker Host gear for ' + str(DockerHostGear.hostname),
            gear_admin_queue='ariane.community.plugin.docker.gears.cache.docker_host_gear@' +
                             str(DockerHostGear.hostname),
            running=False
        )
        self.sleeping_period = config.sleeping_period
        self.service = None
        self.service_name = 'docker_host@'+str(DockerHostGear.hostname)+' gear'
        self.component = DockerComponent.start(
            attached_gear_id=self.gear_id(),
            hostname=DockerHostGear.hostname,
            docker_cli=cli,
            docker_gear_actor_ref=self.actor_ref
        ).proxy()
        self.directory_gear = DirectoryGear.start().proxy()
        self.mapping_gear = MappingGear.start().proxy()

    def synchronize_with_ariane_dbs(self):
        LOGGER.debug("DockerHostGear.synchronize_with_ariane_dbs")
        LOGGER.info("Synchonize with Ariane DBs...")
        self.directory_gear.synchronize_with_ariane_directories(self.component)
        self.mapping_gear.synchronize_with_ariane_mapping(self.component)

    def run(self):
        LOGGER.debug("DockerHostGear.run")
        if self.sleeping_period is not None and self.sleeping_period > 0:
            while self.running:
                time.sleep(self.sleeping_period)
                if self.running:
                    self.component.sniff().get()

    def on_start(self):
        LOGGER.debug("DockerHostGear.on_start")
        self.cache(running=self.running)
        LOGGER.info("Initializing...")
        self.directory_gear.init_ariane_directories(self.component).get()
        self.mapping_gear.init_ariane_mapping(self.component).get()
        self.component.sniff(synchronize_with_ariane_dbs=False).get()
        LOGGER.info("Synchonize with Ariane DBs...")
        self.directory_gear.synchronize_with_ariane_directories(self.component).get()
        self.mapping_gear.synchronize_with_ariane_mapping(self.component).get()
        LOGGER.info("Initialization done.")
        self.running = True
        self.cache(running=self.running)
        self.service = threading.Thread(target=self.run, name=self.service_name)
        self.service.start()

    def on_stop(self):
        LOGGER.debug("DockerHostGear.on_stop")
        try:
            if self.running:
                self.running = False
                self.cache(running=self.running)
            self.service = None
            self.component.stop().get()
            self.directory_gear.stop().get()
            self.mapping_gear.stop().get()
            self.cached_gear_actor.remove().get()
        except Exception as e:
            LOGGER.error(e.__str__())
            LOGGER.error(traceback.format_exc())

    def gear_start(self):
        LOGGER.debug("DockerHostGear.gear_start")
        if self.service is not None:
            LOGGER.warning('docker_host_gear@'+str(DockerHostGear.hostname)+' has been started')
            self.running = True
            self.service = threading.Thread(target=self.run, name=self.service_name)
            self.service.start()
            self.cache(running=self.running)
        else:
            LOGGER.warning('docker_host_gear@' + str(DockerHostGear.hostname) + ' has been restarted')
            self.on_start()

    def gear_stop(self):
        LOGGER.debug("DockerHostGear.gear_stop")
        if self.running:
            LOGGER.warning('docker_host_gear@' + str(DockerHostGear.hostname) + ' has been stopped')
            self.running = False
            self.cache(running=self.running)



# Ariane Docker plugin
# main
#
# Copyright (C) 2016 echinopsii
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import argparse
import json
import logging
import logging.config
import socket
import signal
import time
from ariane_docker.config import Config


__author__ = 'mffrench'

ariane_connector = None
config_path = "/etc/ariane/adocker_configuration.json"
ariane_docker_config = None
docker_host_gear = None


def shutdown_handle(signum, frame):
    LOGGER.info("Ariane Docker@" + socket.gethostname() + " is stopping...")
    if docker_host_gear is not None:
        docker_host_gear.stop().get()
    if ariane_connector is not None:
        time.sleep(5)
        ariane_connector.stop()
    LOGGER.info("Ariane Docker@" + socket.gethostname() + " is stopped...")

parser = argparse.ArgumentParser()
parser.add_argument("-c", "--configuration",
                    help="define your Ariane Docker configuration file path")
args = parser.parse_args()

if args.configuration:
    config_path = args.configuration

try:
    ariane_docker_config = Config().parse(config_path)
except Exception as e:
    print('Ariane Docker plugin config issue: ' + e.__str__())
    exit(1)

try:
    with open(ariane_docker_config.log_conf_file_path, 'rt') as f:
        log_conf = json.load(f)
    logging.config.dictConfig(log_conf)
except Exception as e:
    print("Error while loading configuration file: " + e.__str__())
    logging.basicConfig(format='[%(levelname)s]%(asctime)s - %(name)s - %(message)s', level=logging.WARN)

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)

from ariane_docker.connector import ArianeConnector
from ariane_docker.connector import DockerConnector
from ariane_docker.gears import DockerHostGear

if ariane_docker_config is not None:
    ariane_connector = ArianeConnector(ariane_docker_config)
    docker_connector = DockerConnector(ariane_docker_config)
    if ariane_connector.ready:
        signal.signal(signal.SIGINT, shutdown_handle)
        signal.signal(signal.SIGTERM, shutdown_handle)
        docker_host_gear = DockerHostGear.start(config=ariane_docker_config, cli=docker_connector.cli).proxy()
        LOGGER.info("Ariane Docker@" + socket.gethostname() + " is started...")
        signal.pause()
